#include <ICM42688.h>

void ICM42688::Init(ICM42688_CONFIG icm42688_config){
	// ICM42688 Initialize
	icm42688_config.addr = icm42688_config.addr << 1;
	if(config.timeout <= 0){
		config.timeout = 10000;
	}
	memcpy(&config, &icm42688_config, sizeof(ICM42688_CONFIG));

	// 센서 활성화
	uint8_t data[2] = {ICM42688_PWR_MGMT0, 0x0F};
	HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);

	// Set accelerometer
	data[0] = ICM42688_ACCEL_CONFIG0;
	data[1] = icm42688_config.accel_range | icm42688_config.accel_rate;
	HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);
	data[0] = ICM42688_ACCEL_CONFIG1;
	data[1] = icm42688_config.accel_ui_filt_ord;
	HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);

	// 자이로 센서 설정
	data[0] = ICM42688_GYRO_CONFIG0;
	data[1] = icm42688_config.gyro_range | icm42688_config.gyro_rate;
	HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);
	data[0] = ICM42688_GYRO_CONFIG1;
	data[1] = icm42688_config.gyro_ui_filt_ord;
	HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);

	// Set Low Pass Filter
	data[0] = ICM42688_GYRO_ACCEL_CONFIG0;
	data[1] = icm42688_config.accel_ui_filt_bw | icm42688_config.gyro_ui_filt_bw;
	HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);
}

void ICM42688::Sleep(){

}
void ICM42688::Wakeup(){

}

void ICM42688::setAccelOffset(int16_t offset[]){
	// Select UserBank4
	uint8_t offsetData[5] = {0}, regAddr = 0;
	uint8_t data[2] = {ICM42688_REG_BANK_SEL, ICM42688_USERBANK4};
	HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);

	// Read OFFSET_USER4
	regAddr = ICM42688_OFFSET_USER4;
	HAL_I2C_Master_Transmit(&config.i2c, config.addr, &regAddr, 1, config.timeout);
	HAL_I2C_Master_Receive(&config.i2c, config.addr, &offsetData[0], 1, config.timeout);

	// Assemble Offset Data
	offsetData[0] |= (offset[0] >> 4) & 0xF0;
	offsetData[1] = offset[0] & 0xFF;
	offsetData[2] = offset[1] & 0xFF;
	offsetData[3] = ((offset[1] >> 8) & 0x0F) | ((offset[2] >> 4) & 0xF0);
	offsetData[4] = offset[2] & 0xFF;

	// Write Offset Data
	for(uint8_t i=0;i<5;i++){
		data[0] = ICM42688_OFFSET_USER4+i;
		data[1] = offsetData[i];
		HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);
	}

	// Select UserBank0
	data[0] = ICM42688_REG_BANK_SEL;
	data[1] = ICM42688_USERBANK0;
	HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);
}
void ICM42688::setGyroOffset(int16_t offset[]){
	// Select UserBank4
	uint8_t offsetData[5] = {0};
	uint8_t data[2] = {ICM42688_REG_BANK_SEL, ICM42688_USERBANK4};
	HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);

	// Read OFFSET_USER4
	uint8_t regAddr = ICM42688_OFFSET_USER4;
	HAL_I2C_Master_Transmit(&config.i2c, config.addr, &regAddr, 1, config.timeout);
	HAL_I2C_Master_Receive(&config.i2c, config.addr, &offsetData[4], 1, config.timeout);


	// Assemble Offset Data
	offsetData[0] = offset[0] & 0xFF;
	offsetData[1] = ((offset[1] >> 8 & 0x0F) << 4) | (offset[0] >> 8 & 0x0F);
	offsetData[2] = offset[1] & 0xFF;
	offsetData[3] = offset[2] & 0xFF;
	offsetData[4] |= offset[2] >> 8 & 0x0F;

	// Write Offset Data
	for(uint8_t i=0;i<5;i++){
		data[0] = ICM42688_OFFSET_USER0+i;
		data[1] = offsetData[i];
		HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);
	}

	// Select UserBank0
	data[0] = ICM42688_REG_BANK_SEL;
	data[1] = ICM42688_USERBANK0;
	HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);
}

int16_t *ICM42688::getRawAccel(){
	static int16_t output[3] = {0};
	uint8_t regAddr = ICM42688_ACCEL_DATA_X1, recv[6] = {0};
	uint16_t error = 0;
	error = HAL_I2C_GetState(&config.i2c);
	while(error != HAL_I2C_STATE_READY){};
	error = HAL_I2C_Mem_Read(&config.i2c, config.addr, regAddr, 1, recv, 6, config.timeout);
	if(error != HAL_OK){
		return 0;
	}
	for(uint8_t i=0;i<3;i++){
		output[i] = (recv[2*i] << 8) | recv[2*i + 1];
	}
	return output;
}

int16_t *ICM42688::getRawGyro(){
	static int16_t output[3] = {0};
	uint8_t regAddr = ICM42688_GYRO_DATA_X1, recv[6] = {0};
	uint16_t error = 0;
	error = HAL_I2C_GetState(&config.i2c);
	while(error != HAL_I2C_STATE_READY){};
	error = HAL_I2C_Mem_Read(&config.i2c, config.addr, regAddr, 1, recv, 6, config.timeout);
	if(error != HAL_OK){
		return 0;
	}
	for(uint8_t i=0;i<3;i++){
		output[i] = (recv[2*i] << 8) | recv[2*i + 1];
	}
	return output;
}
int16_t ICM42688::getRawTemp(){
	int16_t output = 0;
	uint8_t regAddr = ICM42688_TEMP_DATA1, recv[2] = {0};
	uint16_t error = 0;
	error = HAL_I2C_GetState(&config.i2c);
	while(error != HAL_I2C_STATE_READY){};
	error = HAL_I2C_Mem_Read(&config.i2c, config.addr, regAddr, 1, recv, 2, config.timeout);
	if(error != HAL_OK){
		return 0;
	}
	output = recv[0] << 8 | recv[1];
	return output;
}

uint8_t ICM42688::getDeviceId(){
	// 디바이스 ID 리턴
	uint8_t regAddr = ICM42688_WHO_AM_I, deviceId = 0;
	uint16_t error = 0;
	error = HAL_I2C_GetState(&config.i2c);
	while(error != HAL_I2C_STATE_READY){};
	error = HAL_I2C_Mem_Read(&config.i2c, config.addr, regAddr, 1, &deviceId, 1, config.timeout);
	if(error != HAL_OK){
		return 0;
	}
	return deviceId;
}


int16_t *ICM42688::getAccelOffset(){
    static int16_t output[3] = {0};

    // Select USERBANK4 and Address
    uint8_t data[2] = {ICM42688_REG_BANK_SEL, ICM42688_USERBANK4};
    HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);

    // Read Offset Register
    uint8_t recvData[5] = {0}, regAddr = ICM42688_OFFSET_USER4;
    HAL_I2C_Master_Transmit(&config.i2c, config.addr, &regAddr, 1, config.timeout);
	HAL_I2C_Master_Receive(&config.i2c, config.addr, recvData, 5, config.timeout);

    // Assemble data
    output[0] = ((recvData[0] & 0xF0) << 4) | recvData[1];
    output[1] = ((recvData[3] & 0x0F) << 8) | recvData[2];
    output[2] = ((recvData[3] & 0xF0) << 4) | recvData[4];
    return output;
}

int16_t *ICM42688::getGyroOffset(){
	static int16_t output[3] = {0};

	// Select USERBANK4 and Address
	uint8_t data[2] = {ICM42688_REG_BANK_SEL, ICM42688_USERBANK4};
	HAL_I2C_Master_Transmit(&config.i2c, config.addr, data, 2, config.timeout);

	// Read Offset Register
	uint8_t recvData[5] = {0}, regAddr = ICM42688_OFFSET_USER0;
	HAL_I2C_Master_Transmit(&config.i2c, config.addr, &regAddr, 1, config.timeout);
	HAL_I2C_Master_Receive(&config.i2c, config.addr, recvData, 5, config.timeout);

	// Assemble data
	output[0] = ((recvData[1] & 0x0F) << 8) | recvData[0];
	output[1] = ((recvData[1] & 0xF0) << 4) | recvData[2];
	output[2] = ((recvData[4] & 0x0F) << 8) | recvData[3];
	return output;
}

int16_t *ICM42688::computeAccelOffset(uint16_t samples){
    uint16_t counter = 0;
    static int16_t output[3] = {0};
    memset(output, 0x00, sizeof(output));
    do{
        int16_t *accel = getRawAccel();
        for(uint8_t i=0;i<3;i++){
            if(counter > 0){
                output[i] = (((float)counter-1.0f)/(float)counter)*(float)output[i] + (1.0f/(float)counter)*(float)accel[i];
            }else{
                output[i] = accel[i];
            }
        }
    }while(++counter<samples);
    return output;
}

int16_t *ICM42688::computeGyroOffset(uint16_t samples){
    uint16_t counter = 0;
    static int16_t output[3] = {0};
    memset(output, 0x00, sizeof(output));
    do{
        int16_t *gyro = getRawGyro();
        for(uint8_t i=0;i<3;i++){
            if(counter > 0){
                output[i] = (((float)counter-1.0f)/(float)counter)*(float)output[i] + (1.0f/(float)counter)*(float)gyro[i];
            }else{
                output[i] = gyro[i];
            }
        }
    }while(++counter<samples);
    return output;
}
