/*
        ICM42688 Library
        Author : Seonguk Jeong
        Created at : 2023-10-1
        Updated at : 2023-10-1
        Target : STM32
*/

#pragma once
#ifndef ROVERDYN_STM32_LIBRARY_SENSORS_ICM422688_H_
#define ROVERDYN_STM32_LIBRARY_SENSORS_ICM42688_H_

#include <stdio.h>
#include <stdint.h>
#include <String.h>
#include <Configure.h>

#define ICM42688_ADDR0		                0x68
#define ICM42688_ADDR1		                0x69

// Register Address
// USER BANK 0
#define ICM42688_DEVICE_CONFIG				0x11
#define ICM42688_DRIVE_CONFIG				0x13
#define ICM42688_INT_CONFIG				0x14
#define ICM42688_ICM42688_FIFO_CONFIG	                0x16
#define ICM42688_TEMP_DATA1				0x1D
#define ICM42688_TEMP_DATA0				0x1E
#define ICM42688_ACCEL_DATA_X1				0x1F
#define ICM42688_ACCEL_DATA_X0				0x20
#define ICM42688_ACCEL_DATA_Y1				0x21
#define ICM42688_ACCEL_DATA_Y0				0x22
#define ICM42688_ACCEL_DATA_Z1				0x23
#define ICM42688_ACCEL_DATA_Z0				0x24
#define ICM42688_GYRO_DATA_X1				0x25
#define ICM42688_GYRO_DATA_X0				0x26
#define ICM42688_GYRO_DATA_Y1				0x27
#define ICM42688_GYRO_DATA_Y0				0x28
#define ICM42688_GYRO_DATA_Z1				0x29
#define ICM42688_GYRO_DATA_Z0				0x2A
#define ICM42688_TMST_FYNCH				0x2B
#define ICM42688_TMST_FYNCL				0x2C
#define ICM42688_INT_STATUS				0x2D
#define ICM42688_FIFO_COUNTH				0x2E
#define ICM42688_FIFO_COUNTL				0x2F
#define ICM42688_FIFO_DATA				0x30
#define ICM42688_APEX_DATA0				0x31
#define ICM42688_APEX_DATA1				0x32
#define ICM42688_APEX_DATA2				0x33
#define ICM42688_APEX_DATA3				0x34
#define ICM42688_APEX_DATA4				0x35
#define ICM42688_APEX_DATA5				0x36
#define ICM42688_INT_STATUS2				0x37
#define ICM42688_INT_STATUS3				0x38
#define ICM42688_SIGNAL_PATH_RESET		        0x4B
#define ICM42688_INTF_CONFIG0				0x4C
#define ICM42688_INTF_CONFIG1				0x4D
#define ICM42688_PWR_MGMT0				0x4E
#define ICM42688_GYRO_CONFIG0				0x4F
#define ICM42688_ACCEL_CONFIG0				0x50
#define ICM42688_GYRO_CONFIG1				0x51
#define ICM42688_GYRO_ACCEL_CONFIG0		        0x52
#define ICM42688_ACCEL_CONFIG1				0x53
#define ICM42688_TMST_CONFIG				0x54
#define ICM42688_APEX_CONFIG0				0x56
#define ICM42688_SMD_CONFIG				0x57
#define ICM42688_FIFO_CONFIG1				0x5F
#define ICM42688_FIFO_CONFIG2				0x60
#define ICM42688_FIFO_CONFIG3				0x61
#define ICM42688_FSYNC_CONFIG				0x62
#define ICM42688_INT_CONFIG0				0x63
#define ICM42688_INT_CONFIG1				0x64
#define ICM42688_INT_SOURCE0				0x65
#define ICM42688_INT_SOURCE1				0x66
#define ICM42688_INT_SOURCE3				0x68
#define ICM42688_INT_SOURCE4				0x69
#define ICM42688_FIFO_LOST_PKT0				0x6C
#define ICM42688_FIFO_LOST_PKT1				0x6D
#define ICM42688_SELF_TEST_CONFIG			0x70
#define ICM42688_WHO_AM_I				0x75
#define ICM42688_REG_BANK_SEL				0x76

// USER BANK 1
#define ICM42688_SENSOR_CONFIG0				0x03
#define ICM42688_GYRO_CONFIG_STATIC2		        0x0B
#define ICM42688_GYRO_CONFIG_STATIC3		        0x0C
#define ICM42688_GYRO_CONFIG_STATIC4		        0x0D
#define ICM42688_GYRO_CONFIG_STATIC5		        0x0E
#define ICM42688_GYRO_CONFIG_STATIC6		        0x0F
#define ICM42688_GYRO_CONFIG_STATIC7		        0x10
#define ICM42688_GYRO_CONFIG_STATIC8		        0x11
#define ICM42688_GYRO_CONFIG_STATIC9		        0x12
#define ICM42688_GYRO_CONFIG_STATIC10	                0x13
#define ICM42688_XG_ST_DATA				0x5F
#define ICM42688_YG_ST_DATA				0x60
#define ICM42688_ZG_ST_DATA				0x61
#define ICM42688_TMSTVAL0				0x62
#define ICM42688_TMSTVAL1				0x63
#define ICM42688_TMSTVAL2				0x64
#define ICM42688_INTF_CONFIG4				0x7A
#define ICM42688_INTF_CONFIG5				0x7B
#define ICM42688_INTF_CONFIG6				0x7C

// USER BANK 2
#define ICM42688_ACCEL_CONFIG_STATIC2	                0x03
#define ICM42688_ACCEL_CONFIG_STATIC3	                0x04
#define ICM42688_ACCEL_CONFIG_STATIC4	                0x05
#define ICM42688_XA_ST_DATA				0x3B
#define ICM42688_YA_ST_DATA				0x3C
#define ICM42688_ZA_ST_DATA				0x3D

// USER BANK 4
#define ICM42688_APEX_CONFIG1				0x40
#define ICM42688_APEX_CONFIG2				0x41
#define ICM42688_APEX_CONFIG3				0x42
#define ICM42688_APEX_CONFIG4				0x43
#define ICM42688_APEX_CONFIG5				0x44
#define ICM42688_APEX_CONFIG6				0x45
#define ICM42688_APEX_CONFIG7				0x46
#define ICM42688_APEX_CONFIG8				0x47
#define ICM42688_APEX_CONFIG9				0x48
#define ICM42688_ACCEL_WOM_X_THR			0x4A
#define ICM42688_ACCEL_WOM_Y_THR			0x4B
#define ICM42688_ACCEL_WOM_Z_THR			0x4C
#define ICM42688_INT_SOURCE6				0x4D
#define ICM42688_INT_SOURCE7				0x4E
#define ICM42688_INT_SOURCE8				0x4F
#define ICM42688_INT_SOURCE9				0x50
#define ICM42688_INT_SOURCE10				0x51
#define ICM42688_OFFSET_USER0				0x77
#define ICM42688_OFFSET_USER1				0x78
#define ICM42688_OFFSET_USER2				0x79
#define ICM42688_OFFSET_USER3				0x7A
#define ICM42688_OFFSET_USER4				0x7B
#define ICM42688_OFFSET_USER5				0x7C
#define ICM42688_OFFSET_USER6				0x7D
#define ICM42688_OFFSET_USER7				0x7E
#define ICM42688_OFFSET_USER8				0x7F

#define ICM42688_USERBANK0				0x00
#define ICM42688_USERBANK1				0x01
#define ICM42688_USERBANK2				0x02
#define ICM42688_USERBANK3				0x03
#define ICM42688_USERBANK4				0x04

#define ICM42688_GYRO_RANGE_2000		(0b000 << 5)
#define ICM42688_GYRO_RANGE_1000		(0b001 << 5)
#define ICM42688_GYRO_RANGE_500		        (0b010 << 5)
#define ICM42688_GYRO_RANGE_250		        (0b011 << 5)
#define ICM42688_GYRO_RANGE_125		        (0b100 << 5)
#define ICM42688_GYRO_RANGE_62_5		(0b101 << 5)
#define ICM42688_GYRO_RANGE_31_25	        (0b110 << 5)
#define ICM42688_GYRO_RANGE_15_625	        (0b111 << 5)
#define ICM42688_GYRO_ODR_32KHZ		        0b0001
#define ICM42688_GYRO_ODR_16KHZ		        0b0010
#define ICM42688_GYRO_ODR_8KHZ		        0b0011
#define ICM42688_GYRO_ODR_4KHZ		        0b0100
#define ICM42688_GYRO_ODR_2KHZ		        0b0101
#define ICM42688_GYRO_ODR_1KHZ		        0b0110
#define ICM42688_GYRO_ODR_500HZ		        0b1111
#define ICM42688_GYRO_ODR_200HZ		        0b0111
#define ICM42688_GYRO_ODR_100HZ		        0b1000
#define ICM42688_GYRO_ODR_50HZ		        0b1001
#define ICM42688_GYRO_ODR_25HZ		        0b1010
#define ICM42688_GYRO_ODR_12_5HZ		0b1011
#define ICM42688_ACCEL_RANGE_16			(0b000 << 5)
#define ICM42688_ACCEL_RANGE_8			(0b001 << 5)
#define ICM42688_ACCEL_RANGE_4			(0b010 << 5)
#define ICM42688_ACCEL_RANGE_2			(0b011 << 5)
#define ICM42688_ACCEL_ODR_32KHZ		0b0001
#define ICM42688_ACCEL_ODR_16KHZ		0b0010
#define ICM42688_ACCEL_ODR_8KHZ		        0b0011
#define ICM42688_ACCEL_ODR_4KHZ		        0b0100
#define ICM42688_ACCEL_ODR_2KHZ		        0b0101
#define ICM42688_ACCEL_ODR_1KHZ		        0b0110
#define ICM42688_ACCEL_ODR_200HZ		0b0111
#define ICM42688_ACCEL_ODR_100HZ		0b1000
#define ICM42688_ACCEL_ODR_50HZ		        0b1001
#define ICM42688_ACCEL_ODR_25HZ		        0b1010
#define ICM42688_ACCEL_ODR_12_5HZ	        0b1011
#define ICM42688_ACCEL_ODR_6_25HZ	        0b1100
#define ICM42688_ACCEL_ODR_3_125HZ	        0b1101
#define ICM42688_ACCEL_ODR_1_5625HZ	        0b1110
#define ICM42688_ACCEL_ODR_500HZ		0b1111
#define ICM42688_ACCEL_UI_FILT_BW0              (0x00)
#define ICM42688_ACCEL_UI_FILT_BW1              (0x01 << 4)
#define ICM42688_ACCEL_UI_FILT_BW2              (0x02 << 4)
#define ICM42688_ACCEL_UI_FILT_BW3              (0x03 << 4)
#define ICM42688_ACCEL_UI_FILT_BW4              (0x04 << 4)
#define ICM42688_ACCEL_UI_FILT_BW5              (0x05 << 4)
#define ICM42688_ACCEL_UI_FILT_BW6              (0x06 << 4)
#define ICM42688_ACCEL_UI_FILT_BW7              (0x07 << 4)
#define CIM42688_ACCEL_UI_FILT_ORD_1ST          0x00
#define CIM42688_ACCEL_UI_FILT_ORD_2ND          (0x01 << 3)
#define CIM42688_ACCEL_UI_FILT_ORD_3RD          (0x02 << 3)
#define ICM42688_ACCEL_DEC2_M2_ORD_3RD          (0b10 << 1)
#define ICM42688_GYRO_UI_FILT_BW0               0x00
#define ICM42688_GYRO_UI_FILT_BW1               0x01
#define ICM42688_GYRO_UI_FILT_BW2               0x02
#define ICM42688_GYRO_UI_FILT_BW3               0x03
#define ICM42688_GYRO_UI_FILT_BW4               0x04
#define ICM42688_GYRO_UI_FILT_BW5               0x05
#define ICM42688_GYRO_UI_FILT_BW6               0x06
#define ICM42688_GYRO_UI_FILT_BW7               0x07
#define ICM42688_GYRO_UI_FILT_ORD_1ST           0b00
#define ICM42688_GYRO_UI_FILT_ORD_2ST           (0b01 << 2)
#define ICM42688_GYRO_UI_FILT_ORD_3ST           (0b10 << 2)
#define ICM42688_GYRO_DEC2_M2_ORD_3RD           0b10
#define ICM42688_TEMP_FILT_BW_4000HZ            0b000
#define ICM42688_TEMP_FILT_BW_170HZ             0b001
#define ICM42688_TEMP_FILT_BW_82HZ              0b010
#define ICM42688_TEMP_FILT_BW_40HZ              0b011
#define ICM42688_TEMP_FILT_BW_20HZ              0b100
#define ICM42688_TEMP_FILT_BW_10HZ              0b101
#define ICM42688_TEMP_FILT_BW_5HZ               0b110

struct ICM42688_CONFIG{
		I2C_HandleTypeDef i2c;
        uint8_t addr;
        uint32_t timeout = 10000;
        uint8_t accel_range;
        uint8_t accel_rate;
        uint8_t accel_ui_filt_ord;
        uint8_t accel_ui_filt_bw;
        uint8_t gyro_range;
        uint8_t gyro_rate;
        uint8_t gyro_ui_filt_ord;
        uint8_t gyro_ui_filt_bw;
        uint8_t temp_filter;
};

class ICM42688{
        private:
			ICM42688_CONFIG config;
        public:
			void Init(ICM42688_CONFIG icm42688_config);
			void Sleep();
			void Wakeup();
			void setAccelOffset(int16_t offset[]);
			void setGyroOffset(int16_t offset[]);
			int16_t *getRawAccel();
			int16_t *getRawGyro();
			int16_t getRawTemp();
			uint8_t getDeviceId();
			int16_t *getAccelOffset();
			int16_t *getGyroOffset();
			int16_t *computeAccelOffset(uint16_t samples);
			int16_t *computeGyroOffset(uint16_t samples);
};
#endif
